import datetime

from sqlalchemy import Column, Integer, String, DateTime

from command_system import RollbotResponse, RollbotFailure, as_plugin, ModelBase, pop_arg


class DnDSession(ModelBase):
    __tablename__ = "dnd_session"
    group_id = Column(String, primary_key=True)
    session_time = Column(DateTime)


class DnDCancellationBlame(ModelBase):
    __tablename__ = "dnd_session_cancel_blame"
    group_id = Column(String, primary_key=True)
    name = Column(String, primary_key=True)
    count = Column(Integer)


WEEKDAYS = {
    "monday": 0,
    "tuesday": 1,
    "wednesday": 2,
    "thursday": 3,
    "friday": 4,
    "saturday": 5,
    "sunday": 6
}


def parse_datetime(arg_text):
    if arg_text is None:
        return

    day_date, arg_text = pop_arg(arg_text)
    day_date = day_date.lower()
    if day_date in WEEKDAYS:
        today = datetime.date.today()
        day_diff = WEEKDAYS[day_date] - today.weekday()
        next_day = today + datetime.timedelta((day_diff) % 7)
        month = next_day.month
        day = next_day.day
        year = next_day.year
    else:
        date_parts = day_date.split("/")
        if len(date_parts) == 3:
            month = date_parts[0]
            day = date_parts[1]
            year = date_parts[2]
        elif len(date_parts) == 2:
            month = date_parts[0]
            day = date_parts[1]
            year = datetime.date.today().year
        else:
            return  # invalid format

    hour = 19
    minute = 45

    if arg_text is None:
        return datetime.datetime(
            int(year), int(month), int(day),
            int(hour), int(minute)
        )

    time, _ = pop_arg(arg_text)
    time_parts = time.split(":")
    hour = time_parts[0]
    if len(time_parts) > 1:
        minute = time_parts[1]

    try:
        ampm = next(args)
    except StopIteration:
        return datetime.datetime(
            int(year), int(month), int(day),
            int(hour), int(minute)
        )

    if ampm.lower() == "pm":
        hour = int(hour) + 12

    return datetime.datetime(
        int(year), int(month), int(day),
        int(hour), int(minute)
    )



def get_latest_session(db, group_id):
    saved = db.query(DnDSession).get(group_id)
    if saved is None:
        return

    if saved.session_time >= datetime.datetime.now():
        return saved

    db.delete(saved)
    return


def get_fmt_next(db, group_id):
    nxt = get_latest_session(db, group_id)
    if nxt is None or nxt.session_time < datetime.datetime.now():
        return "No upcoming session found."

    return "Next DnD session scheduled for " + nxt.session_time.strftime("%b %d, %Y, at %I:%M %p")


def session_next(db, message, arg_text):
    dt = parse_datetime(arg_text)
    if dt is None:
        return RollbotResponse(
            message,
            failure=RollbotFailure.INVALID_ARGUMENTS,
            debugging={"explain": "Could not process argument as a date/time."}
        )

    if dt < datetime.datetime.now():
        return RollbotResponse(
            message,
            failure=RollbotFailure.INVALID_ARGUMENTS,
            debugging={"explain": "Cannot schedule sessions in the past."}
        )

    nxt = get_latest_session(db, message.group_id)
    if nxt is not None:
        db.delete(nxt)

    db.add(DnDSession(group_id=message.group_id, session_time=dt))

    return RollbotResponse(message, txt=get_fmt_next(db, message.group_id))


def session_worst(db, message, arg_text):
    rankings = db.query(DnDCancellationBlame
        ).filter(DnDCancellationBlame.group_id == message.group_id
        ).order_by(DnDCancellationBlame.count.desc()
        ).all()

    if len(rankings) == 0:
        return RollbotResponse(message, txt="I don't have any records of cancelled sessions.")

    rankings = "\n".join(f"{i+1}: {r.name} has cancelled {r.count} session(s)" for i, r in enumerate(rankings))
    return RollbotResponse(message, txt="Rankings:\n"+rankings)


def session_view(db, message, arg_text):
    return RollbotResponse(message, txt=get_fmt_next(db, message.group_id))


def session_late(db, message, arg_text):
    nxt = get_latest_session(db, message.group_id)
    if nxt is None:
        return RollbotResponse(message, txt="No upcoming session found.")

    now = datetime.datetime.now()
    nxt = nxt.session_time
    if nxt >= now:
        return RollbotResponse(message, txt="You're not late yet!")

    late = (now - nxt).total_seconds() / 60
    return RollbotResponse(message, txt="You're running %.2f minutes late..." % late)


def session_cancel(db, message, arg_text):
    nxt = get_latest_session(db, message.group_id)
    if nxt is None:
        return RollbotResponse(message, txt="No upcoming session found.")

    db.delete(nxt)

    if arg_text is not None and len(arg_text) > 0:
        counter = db.query(DnDCancellationBlame).get((message.group_id, arg_text))
        if counter is not None:
            counter.count += 1
        else:
            db.add(DnDCancellationBlame(group_id=message.group_id, name=arg_text, count=1))
        return RollbotResponse(message, txt=f"Sigh, session cancelled, and it's all {arg_text}'s fault!")
    else:
        return RollbotResponse(message, txt=f"Sigh, session cancelled...")


SUBCOMMANDS = dict(
    next=session_next,
    worst=session_worst,
    view=session_view,
    late=session_late,
    cancel=session_cancel
)

NEEDS_ADMIN = ("next", "cancel")


@as_plugin
def session(db, message):
    subcommand, arg_text = pop_arg(message.raw_args)

    if subcommand is None:
        return RollbotResponse(message, failure=RollbotFailure.MISSING_SUBCOMMAND)

    if subcommand not in SUBCOMMANDS:
        return RollbotResponse(
            message,
            failure=RollbotFailure.INVALID_SUBCOMMAND,
            debugging={
                "explain": f"Must be one of: {', '.join(repr(s) for s in SUBCOMMANDS)}."
            }
        )

    if subcommand in NEEDS_ADMIN and not message.from_admin:
        return RollbotResponse(message, failure=RollbotFailure.PERMISSIONS)

    return SUBCOMMANDS[subcommand](db, message, arg_text)
